package keybase

import (
	"bufio"
	"context"
	"fmt"
	"io"
	"sync"

	"git.hugfreevikings.wtf/keybase/keybase/pkg/ctxreader"
)

type ChatAPI struct {
	sync.Mutex
	toStdin    chan []byte
	fromStdout chan msg
	fromListen chan []byte
	kbLoc      string
}

type msg interface{}

type apiMsg []byte

type errMsg struct {
	Err error
}

func (e errMsg) Error() string {
	return e.Err.Error()
}

func NewChatAPI(ctx context.Context, opts Options) (*ChatAPI, error) {
	var err error
	if opts.ChannelBufferSize <= 0 {
		opts.ChannelBufferSize = 10
	}

	// set up the channels
	api := ChatAPI{
		toStdin:    make(chan []byte, opts.ChannelBufferSize),
		fromStdout: make(chan msg, opts.ChannelBufferSize),
		fromListen: make(chan []byte, opts.ChannelBufferSize),
	}

	// get the basics
	if opts.KeybaseLocation == "" {
		opts.KeybaseLocation, err = locateKeybase()
		if err != nil {
			return nil, fmt.Errorf("failed to locate keybase: %v", err)
		}
	}
	api.kbLoc = opts.KeybaseLocation

	// set up the api pipes
	chatAPI, err := newStandardAPICmd(context.TODO(), opts, apiNameChat)
	if err != nil {
		return nil, fmt.Errorf("failed to create chat api command")
	}
	chatListenAPI, err := newListenAPICmd(context.TODO(), opts, apiNameChat)
	if err != nil {
		return nil, fmt.Errorf("failed to create chat api-listen command")
	}
	chatStdin, chatStdout, chatStderr := chatAPI.GetPipes()
	_, chatListenStdout, _ := chatListenAPI.GetPipes()

	// connect the pipes to the channels
	err = connectChatPipes(ctx, &api, chatStdin, chatStdout, chatStderr, chatListenStdout)
	if err != nil {
		return nil, fmt.Errorf("failed to connect chat pipes: %v", err)
	}

	// start the api commands
	err = chatAPI.Start()
	if err != nil {
		return nil, fmt.Errorf("failed to start chat api: %v", err)
	}
	err = chatListenAPI.Start()
	if err != nil {
		return nil, fmt.Errorf("failed to start chat listen-api: %v", err)
	}

	return &api, nil
}

// connectChatPipes creates goroutines to send messages from the keybase binary into the
// appropriate channels to be processed
func connectChatPipes(ctx context.Context, api *ChatAPI, chatStdin io.WriteCloser, chatStdout, chatStderr, chatListenStdout io.ReadCloser) error {
	// listen reader
	go func() {
		cr := ctxreader.NewContextReader(ctx, chatListenStdout)
		scanner := bufio.NewScanner(cr)
		for scanner.Scan() {
			api.fromListen <- scanner.Bytes()
		}
	}()

	// writing to stdin
	go func() {
		for {
			select {
			case msg := <-api.toStdin:
				_, err := chatStdin.Write(msg)
				if err != nil {
					api.fromStdout <- errMsg{Err: err}
				}
			case <-ctx.Done():
				return
			}
		}
	}()

	// reading from stdout
	go func() {
		cr := ctxreader.NewContextReader(ctx, chatStdout)
		scanner := bufio.NewScanner(cr)
		for scanner.Scan() {
			api.fromStdout <- scanner.Bytes()
		}
	}()

	// reading from stderr
	go func() {
		cr := ctxreader.NewContextReader(ctx, chatStderr)
		scanner := bufio.NewScanner(cr)
		for scanner.Scan() {
			api.fromStdout <- errMsg{Err: fmt.Errorf("%v", scanner.Text())}
		}
	}()

	return nil
}

func (c *ChatAPI) SendRaw(ctx context.Context, msg []byte) ([]byte, error) {
	c.Lock()
	defer c.Unlock()

	// send message to api
	c.toStdin <- msg

	// wait for response
	for {
		select {
		case resp := <-c.fromStdout:
			switch resp := resp.(type) {
			case apiMsg:
				return resp, nil
			case errMsg:
				return nil, resp.Err
			}
		case <-ctx.Done():
			return nil, ctx.Err()
		}
	}
}
