package keybase

import (
	"context"
	"errors"
	"fmt"
	"io"
	"os/exec"
)

type apiName string

const (
	apiNameChat    apiName = "chat"
	apiNameKVStore apiName = "kvstore"
	apiNameTeam    apiName = "team"
	apiNameWallet  apiName = "wallet"
)

// apiCommand holds the pipes that connect to a keybase api command
type apiCommand struct {
	stderr    io.ReadCloser
	stdin     io.WriteCloser
	stdout    io.ReadCloser
	startFunc func() error
}

// newAPICmd returns an apiCommand for a standard keybase api command (i.e. not an
// api-listen command)
func newAPICmd(ctx context.Context, kbLoc string, args []string) (*apiCommand, error) {
	var err error

	// create a new apiCommand
	api := apiCommand{}

	// set up the command
	cmd := exec.CommandContext(ctx, kbLoc, args...)

	// hook up the stderr pipe
	if api.stderr, err = cmd.StderrPipe(); err != nil {
		return nil, fmt.Errorf("failed to get pipe: %v", err)
	}

	// hook up the stdout pipe
	if api.stdout, err = cmd.StdoutPipe(); err != nil {
		return nil, fmt.Errorf("failed to get pipe: %v", err)
	}

	// hook up the stdin pipe
	if api.stdin, err = cmd.StdinPipe(); err != nil {
		return nil, fmt.Errorf("failed to get pipe: %v", err)
	}

	// create an anonymous function to start the command
	api.startFunc = func() error {
		if err := cmd.Start(); err != nil {
			return fmt.Errorf("failed to start command: %v", err)
		}
		return nil
	}

	return &api, nil
}

func newStandardAPICmd(ctx context.Context, opts Options, name apiName) (*apiCommand, error) {
	if opts.KeybaseLocation == "" {
		return nil, errors.New("keybase location not set")
	}
	args, err := buildArgs(opts, string(name), "api")
	if err != nil {
		return nil, fmt.Errorf("failed to build args: %v", err)
	}
	return newAPICmd(ctx, opts.KeybaseLocation, args)
}

func newListenAPICmd(ctx context.Context, opts Options, name apiName) (*apiCommand, error) {
	if opts.KeybaseLocation == "" {
		return nil, errors.New("keybase location not set")
	}
	args, err := buildArgs(opts, string(name), "api-listen")
	if err != nil {
		return nil, fmt.Errorf("failed to build args: %v", err)
	}
	return newAPICmd(ctx, opts.KeybaseLocation, args)
}

func (c *apiCommand) Start() error {
	return c.startFunc()
}

func (c *apiCommand) GetPipes() (stdin io.WriteCloser, stdout io.ReadCloser, stderr io.ReadCloser) {
	return c.stdin, c.stdout, c.stderr
}
